"""Definition of the various Sampler objects"""

import time
import copy
import multiprocessing
import queue

import numpy as np

from .calibration import Calibration

class Sampler(object):
    """Base Sampler class

        Attributes:
            cal_width (:obj:`int`): The width of the image(s) for which the calibrations are
            cal_height (:obj:`int`): The height of the image(s) for which the calibrations are
            aspect_ratio (:obj:`float`): The calibration aspect ratio.
    """

    def __init__(self, cal_width, cal_height):
        """Initializes a Sampler object

        Args:
            cal_width (:obj:`int`): The width of the image(s) for which the calibrations are
            cal_height (:obj:`int`): The height of the image(s) for which the calibrations are
        """
        self.cal_width = cal_width
        self.cal_height = cal_height
        self.aspect_ratio = cal_width / cal_height

    def next(self):
        """Should be implemented by children classes"""
        raise NotImplementedError

    def histogram(self, reference, n_samples=1000, n_bins=10, return_values=False, **kwargs):
        """Obtains a histogram of APPD values

        Samples n_samples times from the ranges of the parameters, calculates their
        APPD values with respect to a reference Calibration and builds a histogram from them.

        Args:
            reference (:obj:`Calibration`): A reference Calibration object
            n_samples (:obj:`int`): Number of samples
            n_bins (:obj:`int`): Number of histogram bins
            return_values (:obj:`bool`): If `True`, will also return the sampled APPD values
            **kwargs: Arguments to be passed to the `appd` method of Calibration (width and height required)

        Returns:
            tuple: a float or a tuple containing:

                hist (:obj:`numpy array`): The values of the histogram (number of occurrences)

                bin_edges (:obj:`numpy array`): The bin edges `(length(hist)+1)`

                appd_values (:obj:`list` of :obj:`float`): The sampled values (only if `return_values` is set to `True`)

        """

        APPD_values = list()

        for _ in range(n_samples):
            sample_obtained = False
            while not sample_obtained:
                try:
                    APPD_values.append(self.next().appd(reference=reference, **kwargs))
                    sample_obtained = True
                except RuntimeError as e:
                    # If the parameter set is bad (no valid region or too small) sample again
                    pass
        if return_values:
            r = np.histogram(APPD_values, bins=n_bins)
            return r[0], r[1], APPD_values
        else:
            return np.histogram(APPD_values, bins=n_bins)

class ParallelBufferedSampler(Sampler):
    """ParallelBufferedSampler paralelizes and buffers a given sampler.

    Creates several background treads that sample :class:`~.Calibration` objects and maintain a
    buffer of a specified size.

    In order to stop the threads (and gracefully exit the parent process) you
    have to use the :meth:`~.ParallelBufferedSampler.stop` method.

    """

    def __init__(self, sampler, n_jobs=4, buffer_size=16, cache_size=0):
        """Given an initialized sampler, abstracts its parallelization and buffering.

        Args:
            sampler (:obj:`Sampler`): A Sampler object to parallelize
            n_jobs (:obj:`int`): Number of threads to create
            buffer_size (:obj:`int`): Number of calibrations to keep in the buffer
            cache_size (:obj:`int`): Number of objects to keep in a cache in order to all fast sampling.

        """

        # Copy all the attributes of the sampler to the ParallelBufferedSampler
        self.__dict__ = sampler.__dict__.copy()

        # Store the init arguments
        self.init_sampler = sampler
        self.n_jobs = n_jobs
        self.buffer_size = buffer_size
        self.cache_size = cache_size
        self.cache = list()

        # Make a copy of the sampler for every thread
        self.samplers = list()
        for _ in range(n_jobs): self.samplers.append(copy.deepcopy(sampler))

        # Create the buffer Queue
        self.bufferQueue = multiprocessing.Queue(maxsize=buffer_size)
        self.stopTrigger = multiprocessing.Event()

        # The function that each thread runs
        def thread_fun(sampler, buffer, stop):
            # Make sure each copy has a unique numpy seed
            np.random.seed(multiprocessing.current_process().pid)

            # Keep filling the buffer until requested to stop
            # Convoluted because we want to catch when 'stop' becomes set
            # and at the same time we don't want to sample unnecessarily
            new_sample = None
            while not stop.is_set():
                if new_sample is None:
                    new_sample = sampler.next()
                try:
                    buffer.put(new_sample, block=True, timeout=0.5)
                    new_sample = None
                except queue.Full:
                    pass

        # Start the threads
        self.processes = list()
        for i in range(n_jobs):
            p = multiprocessing.Process(target=thread_fun, args=(self.samplers[i], self.bufferQueue, self.stopTrigger, ))
            p.start()
            self.processes.append(p)

    def next(self):
        """Provide the next :obj:`Calibration` from the buffer.

        Returns:
            :obj:`Calibration`: A :obj:`Calibration` object
        """
        try:
            # Make sure we don't wait only if there's something in the cache to pickup
            if len(self.cache) > 0:
                block = False
            else:
                block = True

            new_sample = self.bufferQueue.get(block=block)

            # Add to the cache if desired:
            # print("cache_size", self.cache_size)
            if self.cache_size > 0:
                # print("ADDING TO CACHE")
                new_sample_cp = copy.deepcopy(new_sample)

                # Delete the maps to preserve memory
                new_sample_cp.delete_maps()
                self.cache.append(new_sample_cp)

                # Maintain the maximum cache size
                while len(self.cache) > self.cache_size:
                    del self.cache[0]

        except queue.Empty: # If we cannot provide a real unique new sample give one from the cache
            new_sample = self.cache[np.random.randint(0, len(self.cache))]

        return new_sample

    def stop(self):
        """Stops the background processes that fill the buffer."""
        self.stopTrigger.set()

        # Give the threads a chance to shut off properly
        time.sleep(1.5)
        for i in range(len(self.processes)):
            self.processes[i].terminate()


class ParameterSampler(Sampler):
    """ParameterSampler provides uniform independent sampling within specified ranges of
    calibration parameters.

    Attributes:
        width (:obj:`int`): The width of the image(s) for which the calibrations are
        height (:obj:`int`): The height of the image(s) for which the calibrations are
        aspect_ratio (:obj:`float`): The calibration aspect ratio.
        ranges (:obj:`dict`): The provided sampling ranges for the calibration parameters
    """

    def __init__(self, ranges, cal_width, cal_height):
        """Initializes a ParameterSampler object

        Args:
            ranges (:obj:`dict`): A dictionary with keys `[fx, fy, cx, cy, k1, k2, p1, p2, k3]` and elements tuples
                describing the sampling range for each parameter. All intrinsic parameters must be provided.
                Missing distortion parameters will be sampled as 0. Parameters can be kept constant if both limits
                are the same.
            cal_width (:obj:`int`): The width of the image(s) for which the calibrations are
            cal_height (:obj:`int`): The height of the image(s) for which the calibrations are
        Raises:
            ValueError: If one of `[fx, fy, cx, cy]` is missing from `ranges`
        """

        super(ParameterSampler, self).__init__(cal_width=cal_width, cal_height=cal_height)

        # Validate the ranges
        for key in ['fx', 'fy', 'cx', 'cy']:
            if key not in ranges: raise ValueError("Key %s missing in ranges" % key)
        for key in ['k1', 'k2', 'p1', 'p2', 'k3']:
            if key not in ranges: ranges[key] = (0, 0)
        self.ranges = ranges

    def next(self):
        """Generator method providing a randomly sampled :obj:`Calibration`

        Returns:
            :obj:`Calibration`: A :obj:`Calibration` object
        """

        # Sample the values
        sample = dict()
        for key in self.ranges:
            sample[key] = np.random.uniform(self.ranges[key][0], self.ranges[key][1])

        # Construct a Calibration object
        K = np.array([[sample['fx'],    0.0,            sample['cx']],
                      [0.0,             sample['fy'],   sample['cy']],
                      [0.0,             0.0,            1.0         ]])
        D = np.array([sample['k1'], sample['k2'], sample['p1'], sample['p2'], sample['k3']])

        return Calibration(K=K, D=D, width=self.cal_width, height=self.cal_height)

class TriangularParameterSampler(Sampler):
    """ParameterSampler provides uniform independent sampling within specified ranges of
    calibration parameters.

    Attributes:
        width (:obj:`int`): The width of the image(s) for which the calibrations are
        height (:obj:`int`): The height of the image(s) for which the calibrations are
        aspect_ratio (:obj:`float`): The calibration aspect ratio.
        ranges (:obj:`dict`): The provided sampling ranges for the calibration parameters
    """

    def __init__(self, ranges, reference, cal_width, cal_height):
        """Initializes a ParameterSampler object

        Args:
            ranges (:obj:`dict`): A dictionary with keys `[fx, fy, cx, cy, k1, k2, p1, p2, k3]` and elements tuples
                describing the sampling range for each parameter. All intrinsic parameters must be provided.
                Missing distortion parameters will be sampled as 0. Parameters can be kept constant if both limits
                are the same.
            reference (:obj:`Calibration`): A reference object that will  be used to determine the mode of the
                triangular distribution for each axis
            cal_width (:obj:`int`): The width of the image(s) for which the calibrations are
            cal_height (:obj:`int`): The height of the image(s) for which the calibrations are
        Raises:
            ValueError: If one of `[fx, fy, cx, cy]` is missing from `ranges`
        """

        super(TriangularParameterSampler, self).__init__(cal_width=cal_width, cal_height=cal_height)

        # Validate the ranges
        for key in ['fx', 'fy', 'cx', 'cy']:
            if key not in ranges: raise ValueError("Key %s missing in ranges" % key)
        for key in ['k1', 'k2', 'p1', 'p2', 'k3']:
            if key not in ranges: ranges[key] = (0, 0)
        self.ranges = ranges

        ref_K = reference.get_K(height=cal_height)
        self.mode = {'fx': ref_K[0,0],
                     'fy': ref_K[1,1],
                     'cx': ref_K[0,2],
                     'cy': ref_K[1,2],
                     'k1': reference.k1,
                     'k2': reference.k2,
                     'p1': reference.p1,
                     'p2': reference.p2,
                     'k3': reference.k3}

    def next(self):
        """Generator method providing a randomly sampled :obj:`Calibration`

        Returns:
            :obj:`Calibration`: A :obj:`Calibration` object
        """

        # Sample the values
        sample = dict()
        for key in self.ranges:
            sample[key] = np.random.triangular(self.ranges[key][0], self.mode[key]+1e-10, self.ranges[key][1]+2e-10)

        # Construct a Calibration object
        K = np.array([[sample['fx'],    0.0,            sample['cx']],
                      [0.0,             sample['fy'],   sample['cy']],
                      [0.0,             0.0,            1.0         ]])
        D = np.array([sample['k1'], sample['k2'], sample['p1'], sample['p2'], sample['k3']])

        return Calibration(K=K, D=D, width=self.cal_width, height=self.cal_height)


class UniformAPPDSampler(Sampler):
    """UniformAPPDSampler provides parameter sampling that results in approximately uniform APPD distribution

        Attributes:
            width (:obj:`int`): The width of the image(s) for which the calibrations are
            height (:obj:`int`): The height of the image(s) for which the calibrations are
            aspect_ratio (:obj:`float`): The calibration aspect ratio.
            ranges (:obj:`dict`): The provided sampling ranges for the calibration parameters
            temperature (:obj:`float`): Temperature used for Gibbs acceptance sampling
            reference (:obj:`Calibration`): The reference :obj:`Calibration`
            bin_edges (:obj:`list` of :obj:`float`): Marks the bin edges
            bin_counts (:obj:`list` of :obj:`int`): Stores how many samples were generated from each bin so far
    """

    def __init__(self, ranges, cal_width, cal_height, reference, temperature=1, appd_range_dicovery_samples=1000,
                 appd_range_bins=10, init_jobs=1, sampler='uniform', **kwargs):
        """Initializes a UniformAPPDSampler object

        Args:
            ranges (:obj:`dict`): A dictionary with keys `[fx, fy, cx, cy, k1, k2, p1, p2, k3]` and elements tuples
                describing the sampling range for each parameter. All intrinsic parameters must be provided.
                Missing distortion parameters will be sampled as 0. Parameters can be kept constant if both limits
                are the same.
            cal_width (:obj:`int`): The width of the image(s) for which the calibrations are
            cal_height (:obj:`int`): The height of the image(s) for which the calibrations are
            reference (:obj:`Calibration`): A reference :class:`~.Calibration` object
            temperature (:obj:`float`): Temperature used for Gibbs acceptance sampling
            appd_range_dicovery_samples (:obj:`int`): Number of samples obtained in order to find the
                range of achievable APPD values
            appd_range_bins (:obj:`int`): Number of histogram bins
            init_jobs (:obj:`int`): Number of jobs used for the initialization sampling
            sampler (:obj:`Sampler`): A type of sampler to use, 'uniform' or 'triangular'
            **kwargs: Arguments to be passed to the appd method of :class:`~.Calibration` (`width` and `height`
                required, optionally `map_width`, `map_height`, `interpolation`)
        Raises:
            ValueError: If one of `[fx, fy, cx, cy]` is missing from ranges
            RuntimeException: If a histogram bin with zero elements is encountered.
        """

        super(UniformAPPDSampler, self).__init__(cal_width=cal_width, cal_height=cal_height)

        self.temperature = temperature
        self.reference = reference
        self.kwargs = kwargs
        self.sampler = sampler
        if self.sampler == 'triangular':
            ref_K = reference.get_K(height=cal_height)
            self.mode = {'fx': ref_K[0,0],
                         'fy': ref_K[1,1],
                         'cx': ref_K[0,2],
                         'cy': ref_K[1,2],
                         'k1': reference.k1,
                         'k2': reference.k2,
                         'p1': reference.p1,
                         'p2': reference.p2,
                         'k3': reference.k3}

        # Validate the ranges
        for key in ['fx', 'fy', 'cx', 'cy']:
            if key not in ranges: raise ValueError("Key %s missing in ranges" % key)
        for key in ['k1', 'k2', 'p1', 'p2', 'k3']:
            if key not in ranges: ranges[key] = (0, 0)
        self.ranges = ranges

        # Find the approximate range of normalized APPD values achievable by the ranges provided by sampling
        ps = ParameterSampler(ranges=self.ranges, cal_width=self.cal_width, cal_height=self.cal_height)
        # Parallelize if requested
        if init_jobs > 1:
            ps = ParallelBufferedSampler(sampler=ps, n_jobs=init_jobs, buffer_size=init_jobs*4)
        hist, bin_edges = ps.histogram(reference=reference, n_samples=appd_range_dicovery_samples,
                                       n_bins=appd_range_bins, normalized=True, **kwargs)
        if init_jobs > 1:
            ps.stop()

        # Make sure that all bins have at least 1 sample. This ensures that all bins can be sampled and the procedure
        # won't end up in a deadlock
        for b_idx, b in enumerate(hist):
            if not b > 0:
                bin_edges = bin_edges[:b_idx+1]

        # Make the attributes that store the bins
        self.APPDrange = (bin_edges[0], bin_edges[-1])
        self.bin_edges = np.array(bin_edges)
        self.bin_edges[0] = -np.inf
        self.bin_edges[-1] = np.inf
        self.bin_counts = np.array([0]*(len(bin_edges)-1))


    def next(self):
        """Generator method providing a randomly sampled Calibration that results in a
        nearly uniform APPD distribution

        Samples are taken uniformly but are accepted with probability `-diff_with_smallest_bin/temperture`.

        Returns:
            :obj:`Calibration`: A :obj:`Calibration` object
        """

        # Calculate the current acceptance probabilities
        accept_probs = np.exp(-(self.bin_counts-np.min(self.bin_counts))/self.temperature)
        # print('accept_probs')
        # for i in range(len(accept_probs)):
        #     print(self.bin_edges[i+1], accept_probs[i], self.bin_counts[i])
        # print('sampled so far', np.sum(self.bin_counts))

        # Sample until we accept
        accepted = False
        while not accepted:

            # Sample the values until a valid rectification is obtained
            sample_obtained = False
            while not sample_obtained:

                sample = dict()
                for key in self.ranges:
                    if self.sampler == 'uniform':
                        sample[key] = np.random.uniform(self.ranges[key][0], self.ranges[key][1])
                    elif self.sampler == 'triangular':
                        sample[key] = np.random.triangular(self.ranges[key][0], self.mode[key]+1e-10, self.ranges[key][1]+2e-10)
                    else:
                        raise Exception('Type of sampler not supported')

                # Construct a Calibration object
                K = np.array([[sample['fx'], 0.0, sample['cx']],
                              [0.0, sample['fy'], sample['cy']],
                              [0.0, 0.0, 1.0]])
                D = np.array([sample['k1'], sample['k2'], sample['p1'], sample['p2'], sample['k3']])

                c = Calibration(K=K, D=D, width=self.cal_width, height=self.cal_height)

                # Calculate the APPD for this calibration
                try:
                    appd = c.appd(self.reference, normalized=True, **self.kwargs)
                    sample_obtained = True
                except RuntimeError as e:
                    # If the parameter set is bad (no valid region or too small) sample again
                    pass

            # Take a chance on accepting
            for i in range(len(self.bin_edges)-1):
                if self.bin_edges[i+1]>appd:
                    appd_in_bin = i
                    break

            if np.random.rand() < accept_probs[appd_in_bin]:
                accepted = True
                self.bin_counts[appd_in_bin] += 1

        return c



"""Definition of the Calibration object"""

import copy

import numpy as np
import cv2

class Calibration(object):
    """The Calibration object represents a set of intrinsic and distortion
    parameters and enables basic computation on them.

    Attributes:
        n_fx (:obj:`float`): Normalized fx intrinsic camera parameter (for height 1 pixel).
        n_fy (:obj:`float`): Normalized fy intrinsic camera parameter (for height 1 pixel).
        n_cx (:obj:`float`): Normalized cx intrinsic camera parameter (for height 1 pixel).
        n_cy (:obj:`float`): Normalized cy intrinsic camera parameter (for height 1 pixel).
        k1 (:obj:`float`): The k1 radial distortion parameter.
        k2 (:obj:`float`): The k2 radial distortion parameter.
        k3 (:obj:`float`): The k3 radial distortion parameter.
        p1 (:obj:`float`): The p1 tangential distortion parameter.
        p2 (:obj:`float`): The p2 tangential distortion parameter.
        aspect_ratio (:obj:`float`): The calibration aspect ratio.

    """

    def __init__(self, K, D, width, height):
        """Initializes a Calibration object.

        Initializes a calibration object from a set of intrinsic parameters
        and distortion coefficients.

        Args:
            K: A list with the camera matrix values ordered as `[fx, fy, cx, cy]` or
                a 3x3 array representing a camera matrix
            D: A list of distortion coefficients ordered as `[k1, k2, p1, p2, k3]`, all 5
                coefficients should be provided
            width (:obj:`int`): Width of the image resolution at which the camera matrix was obtained
            height (:obj:`int`): Height of the image resolution at which the camera matrix was obtained

        Raises:
            TypeError: If any of the inputs do not match the expected types or dimensions.

        """

        # Normalize and save the inputs
        self.aspect_ratio = float(width) / height
        K = np.array(K)
        if len(K.flatten()) == 4:
            fx, fy, cx, cy = (K[0], K[1], K[2], K[3])
        elif K.shape == (3, 3):
            fx, fy, cx, cy = (K[0, 0], K[1, 1], K[0, 2], K[1, 2])
        else:
            raise TypeError("Input K was not a list of length 4 or a 3x3 matrix")

        # Normalize all by the height so that we and with an image with height 1 and
        # width equal to the aspect ratio self.ar
        self.n_fx = float(fx) / height
        self.n_cx = float(cx) / height
        self.n_fy = float(fy) / height
        self.n_cy = float(cy) / height

        D = np.array(D).flatten()
        if len(D) == 5:
            self.k1, self.k2, self.p1, self.p2, self.k3 = (D[0], D[1], D[2], D[3], D[4])
        else:
            raise TypeError("Input D was not a list of length 5")

        # Initialize some properties
        self._size_used_for_calculations = None
        self.new_camera_matrix = None
        self.valid_region = None
        self.undistort_x = None
        self.undistort_y = None

    def __repr__(self):
        str = "Calibration with:\n"
        str+= "    \t[%.2f\t%.2f\t%.2f]\n" % (self.n_fx, 0, self.n_cx)
        str+= "K = \t|%.2f\t%.2f\t%.2f]\n" % (0, self.n_fy, self.n_cy)
        str+= "    \t[%.2f\t%.2f\t%.2f]\n\n" % (0, 0, 1)
        str+= "D = \t[%.2f\t%.2f\t%.2f\t%.2f\t%.2f]" % (self.k1, self.k2, self.p1, self.p2, self.k3)
        return str

    def get_K(self, height=1):
        """Calculates the camera matrix for a particular resolution.

        Calculates the camera matrix for an image with height (in pixels) as supplied,
        and width equal to `self.aspect_ratio*height`.

        Args:
            height (:obj:`int`): An interger representing the desired height (in pixels)
                for the camera matrix
        Returns:
            numpy array: A 3x3 camera matrix for the desired resolution
        """

        fx = self.n_fx * height
        fy = self.n_fy * height
        cx = self.n_cx * height
        cy = self.n_cy * height

        return np.array([[fx , 0.0, cx ],
                         [0.0, fy,  cy ],
                         [0.0, 0.0 ,1.0]])

    def get_D(self):
        """Provides the distortion coefficents.

        Returns:
            numpy array: A list of the 5 distortion coefficients `[k1, k2, p1, p2, k3]`
        """
        return np.array([self.k1, self.k2, self.p1, self.p2, self.k3])

    def undistortion_maps(self, width, height, map_width=None, map_height=None, interpolation=cv2.INTER_LANCZOS4):
        """Calculates the undistortion maps.

        Based on OpenCV's solution:
        https://docs.opencv.org/2.4/doc/tutorials/calib3d/camera_calibration/camera_calibration.html

        Args:
            width (:obj:`int`): Desired width for the input image
            height (:obj:`int`): Desired height for the input image
            map_width (:obj:`int`): Desired width for the undistortion maps,
                default None and will use the input width
            map_height (:obj:`int`): Desired height for the undistortion maps,
                default None and will use the input height
            interpolation: cv2 interpolation method, default is `cv2.INTER_LANCZOS4`

        Returns:
            (numpy array, numpy array): The undistortion maps for the x-axis and y-axis

        Raises:
            ValueError: If only one of `map_width` and `map_height` is `None`
        """

        # Calculate the undistorion maps, only if the previous were for a different resolution
        if (width, height) != self._size_used_for_calculations:
            # Calculate the new camera matrix
            self.new_camera_matrix, self.valid_region = cv2.getOptimalNewCameraMatrix(cameraMatrix=self.get_K(height=height),
                                                                                   distCoeffs=self.get_D(),
                                                                                   imageSize=(width, height),
                                                                                   alpha=1)

            # Calculate the undistorting maps
            self.undistort_x, self.undistort_y = cv2.initUndistortRectifyMap(cameraMatrix=self.get_K(height=height),
                                                                             distCoeffs=self.get_D(),
                                                                             R=np.eye(3),
                                                                             newCameraMatrix=self.new_camera_matrix,
                                                                             size=(width, height),
                                                                             m1type=cv2.CV_32FC1)

            # Update the stored height
            self._size_used_for_calculations = (width, height)


        # Resize them to the target output
        if map_width is None and map_height is None:
            resized_map_x = self.undistort_x
            resized_map_y = self.undistort_y
        elif map_width is not None and map_height is not None:
            resized_map_x = cv2.resize(self.undistort_x, (map_width, map_height), interpolation=interpolation)
            resized_map_y = cv2.resize(self.undistort_y, (map_width, map_height), interpolation=interpolation)
        else:
            raise ValueError("Both map_width and map_height should be None or numbers")


        return resized_map_x, resized_map_y

    def undistortion_maps_preserving(self, width, height, min_cropped_size=None, map_width=None,
                                     map_height=None, interpolation=cv2.INTER_LANCZOS4):
        """Calculates the preserving undistortion maps.

        Calculates an undistorion map that crops only the part of the undistorted image that has no missing
        pixels due to the rectification and that is the same aspect ratio as the provided size. Then it
        resizes it up to the original size.

        Args:
            width (:obj:`int`): The desired width for the input image
            height (:obj:`int`): The desired height for the input image
            min_cropped_size (:obj:`int`): Optional, if provided as a `(width, height)` tuple, will raise `RuntimeError` if the
                cropped image is smaller than this
            map_width (:obj:`int`): The desired width for the undistortion maps,
                if `None` will use the input width
            map_height (:obj:`int`): The desired height for the undistortion maps,
                if `None` will use the input height
            interpolation: cv2 interpolation method

        Returns:
            (numpy array, numpy array): The preserving undistortion maps for the x-axis and y-axis

        Raises:
            RuntimeError: If the region of validity is only zeros, if the cropped image
                is smaller than `min_cropped_size`, or if the aspect ratio cannot be preserved
            ValueError: If only one of `map_width` and `map_height` is None
        """

        # Make sure the maps are calculated for the right resolution
        self.undistortion_maps(width=width, height=height)

        # Make sure that the OpenCV valid region is actually valid
        if np.allclose(self.valid_region, 0, rtol=1e-04, atol=1e-07):
            raise RuntimeError("The region of validity is zero")


        # Find how to crop the new image in order to use only the region of correct
        # rectification while maintaining the original aspect ratio
        target_aspect_ratio = float(width) / float(height)

        valid_dict = {'x': (self.valid_region[0], self.valid_region[0] + self.valid_region[2]),
                    'y': (self.valid_region[1], self.valid_region[1] + self.valid_region[3])}
        valid_aspect_ratio = float(valid_dict['x'][1] - valid_dict['x'][0]) / (valid_dict['y'][1] - valid_dict['y'][0])

        # If the image is taller than it should, cut its legs and head
        if valid_aspect_ratio < target_aspect_ratio:
            desired_number_of_rows = int(round((valid_dict['x'][1] - valid_dict['x'][0]) / target_aspect_ratio))
            cut_top = int(((valid_dict['y'][1] - valid_dict['y'][0]) - desired_number_of_rows) / 2.0)
            cropped_region = {
                'x': valid_dict['x'],
                'y': (valid_dict['y'][0] + cut_top, valid_dict['y'][0] + cut_top + desired_number_of_rows)
            }
        else:
            desired_number_of_cols = int(round((valid_dict['y'][1] - valid_dict['y'][0]) * target_aspect_ratio))
            cut_left = int(((valid_dict['x'][1] - valid_dict['x'][0]) - desired_number_of_cols) / 2.0)
            cropped_region = {
                'x': (valid_dict['x'][0] + cut_left, valid_dict['x'][0] + cut_left + desired_number_of_cols),
                'y': valid_dict['y']
            }

        # Make sure the cropped aspect ratio is the same as the target one
        if cropped_region['x'][1] - cropped_region['x'][0] < 10 or (cropped_region['y'][1] - cropped_region['y'][0]) < 10:
            raise RuntimeError("The cropped region is too small: %s" % str(cropped_region))

        cropped_aspect_ratio = float(cropped_region['x'][1] - cropped_region['x'][0]) / (cropped_region['y'][1] - cropped_region['y'][0])
        if np.abs(cropped_aspect_ratio-target_aspect_ratio)>0.1:
            raise RuntimeError("The aspect ratio could not be preserved. Target: %.05f, Cropped: %.05f" %
                               (target_aspect_ratio, cropped_aspect_ratio))

        # Check if the cropped image fits the minimum requirements
        if min_cropped_size is not None and \
            (cropped_region['x'][1] - cropped_region['x'][0] < min_cropped_size[0] \
            or cropped_region['y'][1] - cropped_region['y'][0] < min_cropped_size[1]):
            raise RuntimeError("The cropped image has dimensions %dx%d, when the minimum required is %dx%d" % \
                               (int(cropped_region['x'][1] - cropped_region['x'][0]), int(cropped_region['y'][1] - cropped_region['y'][0]),
                                min_cropped_size[0], min_cropped_size[1]))

        # Set the output resolution
        if map_width is None and map_height is None:
            output_resolution = (width, height)
        elif map_width is not None and map_height is not None:
            output_resolution = (map_width, map_height)
        else:
            raise ValueError("Both map_width and map_height should be None or numbers")

        # Crop and resize the undistortion maps
        # TODO must check if the inputs and the maps are in the dimensions I expect them to be
        preserving_undistort_x = copy.deepcopy(self.undistort_x)
        preserving_undistort_y = copy.deepcopy(self.undistort_y)
        preserving_undistort_x = preserving_undistort_x[cropped_region['y'][0]:cropped_region['y'][1],
                                 cropped_region['x'][0]:cropped_region['x'][1]]
        preserving_undistort_x = cv2.resize(preserving_undistort_x, output_resolution, interpolation=interpolation)
        preserving_undistort_y = preserving_undistort_y[cropped_region['y'][0]:cropped_region['y'][1],
                                 cropped_region['x'][0]:cropped_region['x'][1]]
        preserving_undistort_y = cv2.resize(preserving_undistort_y, output_resolution, interpolation=interpolation)

        return preserving_undistort_x, preserving_undistort_y

    def rectify(self, image, interpolation=cv2.INTER_LANCZOS4, quiet=False, strict=False, result_width=None,
                result_height=None, mode='standard', min_cropped_size=None):
        """Rectify an image.

        Args:
            image (:obj:`numpy array`): The input image. Should be `cv2`-compatible. Should be `[width, height, channels]`
            interpolation: cv2 interpolation method
            quiet (:obj:`bool`): If `True`, won't show a warning if the aspect ratio of the image is not the same
                as the one of the calibration given during the initialization
            strict (:obj:`bool`): If `True`, will raise a `ValueError` exception if there is aspect ratio mismatch
            result_width (:obj:`int`): Desired width for the rectified image, if `None` will use the input height
            result_height (:obj:`int`): Desired height for the rectified image, if `None` will use the input height
            mode (:obj:`str`): Can be `standard` or `preserving`, chooses between standard undistortion (with `undistortion_maps`) and
                undistortion with cropped and rescaled vaild aspect-ratio-preserving region
                (with `undistortion_maps_preserving`).
            min_cropped_size (:obj:`tuple` of :obj:`int`): Passed to `undistortion_maps_preserving` if
                `mode` is set to `preserving`

        Returns:
            numpy array: The resulting image after rectification

        Raises:
            ValueError: See information about the `strict` parameter. Also raised if only one of `map_width` and
                `map_height` is `None`
        """
        image = np.array(image)
        image_height = image.shape[0]
        image_width = image.shape[1]
        image_ar = float(image_width)/image_height

        # Check if the aspect ratios match
        if not quiet and not np.isclose(image_ar, self.aspect_ratio, rtol=1e-04, atol=1e-07):
            msg = "The aspect ratio of the image is not the same as the " + \
                  "one of the calibration given during the initialization"
            if strict: raise ValueError(msg)
            else: print("WARNING: " + msg)

        # Set the output resolution
        if result_width is None and result_height is None:
            output_width, output_height = (image_width, image_height)
        elif result_width is not None and result_height is not None:
            output_width, output_height = (result_width, result_height)
        else:
            raise ValueError("Both image_width and image_height should be None or numbers")

        # Rectify with the desired function
        if mode == 'standard':
            map_x, map_y = self.undistortion_maps(height=image_height, width=image_width, interpolation=interpolation,
                                                  map_width=output_width, map_height=output_height)
        elif mode == 'preserving':
            map_x, map_y = self.undistortion_maps_preserving(height=image_height, width=image_width,
                                                             min_cropped_size=min_cropped_size,
                                                             interpolation=interpolation,
                                                             map_width=output_width, map_height=output_height)
        else:
            raise ValueError("Mode %s is not recognized." % str(mode))

        remappedIm = cv2.remap(image, map_x, map_y, interpolation)

        return remappedIm

    def appd(self, reference, width, height, interpolation=cv2.INTER_LANCZOS4,
             min_cropped_size=None, return_diff_map=False, map_width=None, map_height=None, normalized=False):
        """Calculate the Average Pixel Position Difference.

        Args:
            reference (:obj:`Calibration`): Another Calibration object relative to which the metric will be computed
            width (:obj:`int`): Desired width for the rectified image, a base for calculating the metric
            height (:obj:`int`): Desired height for the rectified image, a base for calculating the metric
            interpolation: `cv2` interpolation method
            quiet (:obj:`bool`): If `True`, won't show a warning if the aspect ratio of the image is not the same
                as the one of the calibration given during the initialization
            strict (:obj:`bool`): If `True`, will raise a `ValueError` exception if there is aspect ratio mismatch
            min_cropped_size (:obj:`tuple` of :obj:`int`): Passed to `undistortion_maps_preserving`
            return_diff_map (:obj:`bool`): If `True`, returns the difference map
            map_width (:obj:`int`): Width for the undistortion map, if `None`, `width`is used. Use the same aspect ratio
                for the maps as for the rectified image
            map_height (:obj:`int`): Height for the undistortion map, if `None`, `height` is used. Use the same aspect ratio
                for the maps as for the rectified image
            normalized (:obj:`bool`): If `True`, normalizes by the map diagonal

        Returns:
            tuple: a float or a tuple containing:

                appd (:obj:`float`): The Average Pixel Position Difference between this calibration and the reference,
                calculated for the provided resolution.

                diff_map (:obj:`numpy array`): If `return_diff_map` is `True` also returns the difference map

        Raises:
            ValueError: See information about the `strict` parameter
            RuntimeError: See information for `undistortion_maps_preserving`

        """

        if map_height == None: map_height=height
        if map_width == None: map_width=width

        # Obtain the rectification maps for this and the reference
        this_map_x, this_map_y = self.undistortion_maps_preserving(width=width, height=height,
                                                                   min_cropped_size=min_cropped_size,
                                                                   interpolation=interpolation,
                                                                   map_width=map_width,
                                                                   map_height=map_height)

        ref_map_x, ref_map_y = reference.undistortion_maps_preserving(width=width, height=height,
                                                                      min_cropped_size=min_cropped_size,
                                                                      interpolation=interpolation,
                                                                      map_width=map_width,
                                                                      map_height=map_height)

        # Calculate the difference maps
        diff_map_x = this_map_x-ref_map_x
        diff_map_y = this_map_y-ref_map_y
        diff_map = np.sqrt(np.square(diff_map_x)+np.square(diff_map_y))

        # Find the mean
        appd = np.mean(diff_map)

        # Apply normalization if requested
        if normalized:
            appd /= np.sqrt(width**2 + height**2)

        if return_diff_map:
            return appd, diff_map
        else:
            return appd

    def delete_maps(self):
        """
        Deletes the stored maps in order to reduce the memory footprint of the object.
        """

        self._size_used_for_calculations = None
        self.new_camera_matrix = None
        self.valid_region = None
        self.undistort_x = None
        self.undistort_y = None



from setuptools import setup

version='0.6.2'

setup(name='carnivalmirror',
      version=version,
      description='A tool to handle and simulate pinhole camera miscalibrations',
      url='https://github.com/AleksandarPetrov/carnivalmirror',
      author='Aleksandar Petrov',
      author_email='alpetrov@ethz.ch',
      license='MIT',
      packages=['carnivalmirror'],
      zip_safe=False,
      install_requires=['numpy>=1.14.0', 'opencv-python>=4.1.0.0', 'matplotlib>=2.2.2'],
      package_data={
            'carnivalmirror': ['test_image_*.png'],
      },
      test_suite='carnivalmirror.tests'
      )
